# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: account.proto
# Protobuf Python Version: 4.25.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\raccount.proto\x12\x07\x61\x63\x63ount\x1a\x1bgoogle/protobuf/empty.proto\"C\n\x04User\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x10\n\x08username\x18\x02 \x01(\t\x12\r\n\x05\x65mail\x18\x03 \x01(\t\x12\x0e\n\x06groups\x18\x04 \x03(\x05\"\x11\n\x0fUserListRequest\"!\n\x13UserRetrieveRequest\x12\n\n\x02id\x18\x01 \x01(\x05\x32\x88\x02\n\x0eUserController\x12\x33\n\x04List\x12\x18.account.UserListRequest\x1a\r.account.User\"\x00\x30\x01\x12(\n\x06\x43reate\x12\r.account.User\x1a\r.account.User\"\x00\x12\x39\n\x08Retrieve\x12\x1c.account.UserRetrieveRequest\x1a\r.account.User\"\x00\x12(\n\x06Update\x12\r.account.User\x1a\r.account.User\"\x00\x12\x32\n\x07\x44\x65stroy\x12\r.account.User\x1a\x16.google.protobuf.Empty\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'account_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_USER']._serialized_start=55
  _globals['_USER']._serialized_end=122
  _globals['_USERLISTREQUEST']._serialized_start=124
  _globals['_USERLISTREQUEST']._serialized_end=141
  _globals['_USERRETRIEVEREQUEST']._serialized_start=143
  _globals['_USERRETRIEVEREQUEST']._serialized_end=176
  _globals['_USERCONTROLLER']._serialized_start=179
  _globals['_USERCONTROLLER']._serialized_end=443
# @@protoc_insertion_point(module_scope)
